#ifndef _H_APPLICATION
#define _H_APPLICATION

#include "common.h"
#include "redclient.h"
#include "redkey.h"
#include "spice_configuration.h"
#include "controller.h"
#include "platform.h"
#include "foreign_menu.h"

class RedScreen;
class Application;
class ScreenLayer;
class GUILayer;
class InputsHandler;
class Monitor;
#ifdef WIN32
class CSpiceXIPC;
#endif

class Event {
public:
    Event() : _refs (1) {}

    virtual void responce(Application& application) = 0;

    Event* ref() { ++_refs; return this;}
    void unref() {if (--_refs == 0) delete this;}

protected:
    virtual ~Event() {}

    boost::detail::atomic_count _refs;
    friend class Application;
    uint32_t _generation;
};

class SyncEvent : public Event {
public:
    SyncEvent();

    void wait();
    bool success() { return !_err;}

    virtual void do_responce(Application& application) {}

protected:
    virtual ~SyncEvent();

private:
    virtual void responce(Application& application);

private:
    Mutex _mutex;
    Condition _condition;
    bool _err;
    bool _ready;
};

class ConnectedEvent : public Event {
public:
    ConnectedEvent() : Event() {}
    virtual void responce(Application& application);
};

class DisconnectedEvent : public Event {
public:
    DisconnectedEvent() : Event() {}
    virtual void responce(Application& application);
};

class CoonnectionError : public Event {
public:
    CoonnectionError(int error_code) : Event(), _error_code (error_code) {}

    virtual void responce(Application& application);
private:
    int _error_code;
};

class ErrorEvent : public Event {
public:
    ErrorEvent() : Event() {}
    virtual void responce(Application& application);
};

struct MonitorInfo {
    int depth;
    Point size;
    Point position;
};

class MonitorsQuery : public SyncEvent {
public:
    MonitorsQuery() {}

    virtual void do_responce(Application& application);

    std::vector<MonitorInfo>& get_monitors() {return _monitors;}

private:
    std::vector<MonitorInfo> _monitors;
};

struct KeyInfo {
    uint32_t _make;
    uint32_t _break;
    bool press;
};

typedef struct StickyInfo {
    bool trace_is_on;
    bool sticky_mode;
    bool key_first_down; // True when (1) a potential sticky key is pressed, 
                          // and none of the other keys are pressed and (2) in the moment 
                          // of pressing, _sticky_mode is false. When the key is up
                          // for the first time, it is set to false.
    bool key_down;       // The physical state of the sticky key. Valid only till
                          // stickiness is removed.
    RedKey key;          // the key that is currently being traced, or, 
                          // if _sticky mode is on, the sticky key
    TimerID timer;
} StickyInfo;

class Application : public Platform::EventListener,
                    public Platform::DisplayModeListner,
                    public ForeignMenuInterface {
public:
    Application();
    virtual ~Application();

    void config(SpiceConfiguration& spice_config);
    static void init();
    static void cleanup();

    int run();
    void quit(int exit_code);
    void push_event(Event* event);
    void set_inputs_handler(InputsHandler& handler);
    void remove_inputs_handler(InputsHandler& handler);
    RedScreen* find_screen(int id);
    RedScreen* get_screen(int id);

    void on_screen_destroyed(int id, bool was_captured);
    void on_mouse_motion(int dx, int dy, int buttons_state);
    void on_mouse_position(int x, int y, int buttons_state, int display_id);
    void on_mouse_down(int button, int buttons_state);
    void on_mouse_up(int button, int buttons_state);
    void on_key_down(RedKey key);
    void on_key_up(RedKey key);
    void on_deactivate_screen(RedScreen* screen);
    void on_activate_screen(RedScreen* screen);
    void on_start_screen_key_interception(RedScreen* screen);
    void on_stop_screen_key_interception(RedScreen* screen);
    virtual void on_app_activated();
    virtual void on_app_deactivated();
    virtual void on_monitors_change();
    virtual void on_display_mode_change();
    void on_connected();
    void on_disconnecting();

    bool rearrange_monitors(RedScreen& screen);
    void enter_full_screen();
    void exit_full_screen();
    bool toggle_full_screen();
    void minimize();
    void show_splash(int screen_id);
    void hide_splash(int screen_id);
    void set_title(std::wstring& title);
    void update_title();
    void hide();
    void show();
    void external_show();
    void connect();
    void set_peer_connect_options(peer_connect_options configuration[RED_CHANNEL_END]);
    void set_controller(SpiceController *controller);
    void set_hotkeys(const SpiceConfiguration::hotkeys_t& hotkeys);
    void set_send_cad(bool sendcad);
    uint32_t get_mouse_mode();
    std::string get_canvas_type();

    void on_foreign_menu_message(FrgMenuMsg *hdr, int32_t opaque_conn_ref);
    void on_foreign_menu_command(int32_t opaque_conn_ref, int32_t id);
    void delete_menu(int32_t opaque_conn_ref);
#ifdef WIN32
    void on_sys_command(WPARAM wparam, LPARAM lparam);
    void create_menu(const wchar_t *resource);
    void delete_menu();
    void set_no_taskmgr(bool no_taskmgr);
    void set_ipc_interface(CSpiceXIPC * pIPCInterface);
#endif

private:
    void process_events();
    int message_loop();
    void abort();
    void init_scan_code(int index);
    void init_korean_scan_code(int index);
    void init_escape_scan_code(int index);
    void init_pause_scan_code();
    void init_key_table();
    uint32_t get_make_scan_code(RedKey key);
    uint32_t get_break_scan_code(RedKey key);
    void unpress_all();
    bool release_capture();
    bool do_connect();
    bool do_disconnect();

    Monitor* find_monitor(int id);
    Monitor* get_monitor(int id);
    void init_monitors();
    void destroy_monitors();
    void assign_monitors();
    void restore_monitors();
    void prepare_monitors();
    void position_screens();
    void show_full_screen();
    void send_key_down(RedKey key);
    void send_key_up(RedKey key);
    void send_alt_ctl_del();
    void send_ctrl_alt_end();
    void send_action_hotkeys(int action);
    void send_hotkeys(const SpiceConfiguration::hotkey_seq_t& hotkeys);
    void menu_item_callback(unsigned int item_id);
    int is_hotkeys_pressed();
    bool is_action_keys_pressed(const SpiceConfiguration::hotkey_seq_t& keys);
    bool is_cad_pressed();
    void do_on_key_up(RedKey key);

    void set_ssl_config(SpiceConfiguration& spice_config);

    friend class MonitorsQuery;
    friend class AutoAbort;

private:
    // define an action function that will be called as a response to user input.
    typedef bool (Application::*action_t)();
    static bool is_sticky_trace_key(RedKey key);
    void reset_sticky();

    // returns the press value before operation (i.e., if it was already pressed)
    bool press_key(RedKey key); 
    bool unpress_key(RedKey key); 

    static void sticky_timeout_proc(void* opaque, TimerID timer);
    void sync_keyboard_modifiers();
private:
    SpiceConfiguration _config;
    RedClient _client;
    peer_connect_options _peers_configuration[RED_CHANNEL_END];
    std::vector<RedScreen*> _screens;
    std::list<Event*> _events;
    RedScreen* _main_screen;
    Mutex _events_lock;
    bool _quitting;
    bool _active;
    bool _full_screen;
    bool _changing_screens;
    bool _send_cad;
    int _exit_code;
    uint32_t _events_gen;
    RedScreen* _active_screen;
    KeyInfo _key_table[REDKEY_NUM_KEYS];
    int _num_keys_pressed;
    SpiceConfiguration::hotkeys_t _hotkeys;
    std::map<int, action_t> _actions;
    std::auto_ptr<GUILayer> _gui_layer;
    InputsHandler* _inputs_handler;
    const MonitorsList* _monitors;
    std::wstring _title;
    std::auto_ptr<SpiceController> _external_controller;
    ForeignMenu _foreign_menu;
    bool _splash_mode;
    bool _sys_key_intercept_mode;
    StickyInfo _sticky_info;
    bool _controller_is_up;
#ifdef WIN32
    CSpiceXIPC* _spice_ipc;
    std::wstring _menu_resource;
    bool _no_taskmgr;
#endif
    std::string _canvas_type;
};

#endif
