
#include "common.h"

#include <boost/program_options.hpp>
#include <vector>
#include <iostream>

#include "command_line.h"
#include "debug.h"

using namespace boost::program_options;
using namespace std;

#define RED_OPTION_SPICE_SERVER "spice-server"
#define RED_OPTION_FULL_SCREEN  "full-screen"
#define RED_OPTION_CONTROLLER   "controller"
#define RED_OPTION_LOGGER_INI   "logger-ini"
#define RED_OPTION_SSL          "ssl"
#define RED_OPTION_SSL_CHANNELS "ssl-channels"
#define RED_OPTION_NO_AUDIO     "no-audio"
#define RED_OPTION_PASSWORD     "password"
#define RED_OPTION_AUTO_RES     "auto-display-res"
#define RED_OPTION_HOTKEY       "hot-key"
#define RED_OPTION_SEND_CAD     "send-ctrl-alt-del"
#define RED_OPTION_CANVAS_TYPE  "canvas_type"
#define RED_OPTION_CA_FILE      "ca-file"
#define RED_OPTION_HOST_SUBJECT "host-subject"
#define RED_OPTION_HELP         "help"

#ifdef WIN32
#define RED_OPTION_NO_TASKMGR   "no-taskmgr"
#endif

#define RED_SHORT_FULL_SCREEN  ",f"
#define RED_SHORT_SSL          ",s"
#define RED_SHORT_NO_AUDIO     ",q"
#define RED_SHORT_PASSWORD     ",p"
#define RED_SHORT_HELP         ",h"
#define RED_SHORT_HOTKEY       ",k"

#define RED_PARAMETER_HOST_INDEX 0
#define RED_PARAMETER_PORT_INDEX 1
#define RED_PARAMETER_SPORT_INDEX 2


const char *red_channel_names[] = { "dummy", "main", "display", "inputs", "cursor", "playback", "record" };

CommandLine::CommandLine(SpiceConfiguration& spice_config) :
    _spice_config(spice_config)
{

}

bool CommandLine::parse_cmd_line(const std::vector<std::string>& args)
{
    std::string encrypt_channels;
    std::string hotkeys;
    std::string host_subject;

    options_description visible("spicec host port [secure-port]", 100);
    visible.add_options()
        (RED_OPTION_FULL_SCREEN RED_SHORT_FULL_SCREEN, "Start the client in a full screen mode.")
        (RED_OPTION_CONTROLLER, "Enable external controller")
        (RED_OPTION_LOGGER_INI, value<std::string>(&_spice_config._logger_ini), "Use this logger configuration file")
        (RED_OPTION_PASSWORD RED_SHORT_PASSWORD, value<std::string>(&_spice_config._password), "Set a ticketing password.")
        (RED_OPTION_SSL RED_SHORT_SSL, value<std::string>(&_spice_config._cipher), "Enable SSL.")
        (RED_OPTION_SSL_CHANNELS, value<std::string>(&encrypt_channels), "Choose which channels to encrypt.")
        (RED_OPTION_NO_AUDIO RED_SHORT_NO_AUDIO, "Disable audio.")
        (RED_OPTION_AUTO_RES, "Automatically configure remote display settings.")
        (RED_OPTION_HOTKEY RED_SHORT_HOTKEY, value<std::string>(&hotkeys), "Set a hot-key combination which release mouse and keyboard.")
        (RED_OPTION_SEND_CAD, "Send the Ctrl-Alt-Delete keys combination.")
#ifdef WIN32
        (RED_OPTION_NO_TASKMGR, "Disable the execution of the Windows Task Manager application.")
#endif
        (RED_OPTION_CANVAS_TYPE, value<std::string>(&_spice_config._canvas_type), "Set the screen rendering type")
        (RED_OPTION_CA_FILE, value<std::string>(&_spice_config._CA_file), "CA file")
        (RED_OPTION_HOST_SUBJECT, value<std::string>(&host_subject), "host certifcate subject")
        (RED_OPTION_HELP RED_SHORT_HELP, "Show usage.")
        ;

    // The following options are not included in the usage string.
    options_description hidden;
    hidden.add_options()
        (RED_OPTION_SPICE_SERVER, value< std::vector<std::string> >(), "")
        ;

    options_description opts;
    opts.add(visible);
    opts.add(hidden);

    positional_options_description popts;
    popts.add(RED_OPTION_SPICE_SERVER, 3);

    variables_map vars;
    store(command_line_parser(args).options(opts).positional(popts).run(), vars);
    notify(vars);

    if (vars.count(RED_OPTION_HELP)) {
        std::cout << std::endl << visible << std::endl;
        return false;
    }

    if (vars.count(RED_OPTION_FULL_SCREEN)) {
        _spice_config._full_screen = true;
        LOG_DEBUG("Opening in a full screen mode.");
    }

    if (!encrypt_channels.empty())
    {
		if (encrypt_channels.compare("all") == 0) {
			memset(_spice_config._encrypt_channels, true, sizeof(_spice_config._encrypt_channels));
		} else {
			for (int i = 0; i < sizeof(red_channel_names) / sizeof(red_channel_names[0]); ++i) {
				_spice_config._encrypt_channels[i] = (encrypt_channels.find(red_channel_names[i]) != std::wstring::npos);
			}
		}

        // SSL cipher must have a value if a channel need to be encrypted.
        if (_spice_config._cipher.empty()) {
            _spice_config._cipher = "DEFAULT";
        }
    }

    if (!_spice_config._cipher.empty()) {
        LOG_DEBUG("SSL is enabled, cipher = " << _spice_config._cipher.c_str());
    }

    if (vars.count(RED_OPTION_CONTROLLER)) {
        _spice_config._init_controller = true;
        LOG_DEBUG("External controller is enabled");
    }

    if (vars.count(RED_OPTION_NO_AUDIO)) {
        _spice_config._no_audio = true;
        LOG_DEBUG("Audio is disabled.");
    }

    if (vars.count(RED_OPTION_SEND_CAD)) {
        _spice_config._send_cad = true;
        LOG_DEBUG("Sending Ctrl-Alt-Delete.");
    }

#ifdef WIN32
    if (vars.count(RED_OPTION_NO_TASKMGR)) {
        _spice_config._no_taskmgr = true;
        LOG_DEBUG("Task manager execution is disabled.");
    }
#endif

    if (!_spice_config._password.empty()) {
        LOG_DEBUG("Ticketing password is set");
    }

    if (vars.count(RED_OPTION_AUTO_RES)) {
        _spice_config._auto_display_res = true;
    }

    if (!hotkeys.empty()) {
        _spice_config.parse_hotkeys(hotkeys);
    }

    if (!host_subject.empty()) {
        _spice_config.parse_host_cert_subject(host_subject);
    }

    if (vars.count(RED_OPTION_SPICE_SERVER))
    {
        const std::vector<std::string>& params =
            vars[RED_OPTION_SPICE_SERVER].as< std::vector<std::string> >();

        if (params.size() > 1) {
            _spice_config._host = params[RED_PARAMETER_HOST_INDEX];
            _spice_config._port = atoi(params[RED_PARAMETER_PORT_INDEX].c_str());
        }
        
        if (params.size() > 2) {
            _spice_config._sport = atoi(params[RED_PARAMETER_SPORT_INDEX].c_str());
        }
    }
    
    if (_spice_config._host.empty() || (_spice_config._port == 0)) {
        throw boost::program_options::error("either host or port are missing");
    }

    return true;
}

#ifdef WIN32
bool CommandLine::parse_cmd_line(const char* cmdline)
{
    std::vector<std::string> args = split_winmain(cmdline);
    return parse_cmd_line(args);
}
#endif

bool CommandLine::parse_cmd_line(int argc, const char** argv)
{
    std::vector<std::string > args;
    const char** now = argv + 1;
    const char** end = argv + argc;

    for (; now < end; ++now) {
        args.push_back(*now);
    }
    return parse_cmd_line(args);
}
