
#include "common.h"

#include <boost/algorithm/string.hpp>
#include <boost/tokenizer.hpp>

#include "spice_configuration.h"
#include "debug.h"
#include "utils.h"

using namespace std;

SpiceConfiguration::SpiceConfiguration () 
    : _port (0)
    , _sport (0)
    , _full_screen (false)
    , _init_controller (false)
    ,_no_audio (false)
    , _auto_display_res (false)
    , _title (L"SPICEc:%d")
    , _logger_ini ("logger.ini")
    , _send_cad (false)
#ifdef WIN32
    , _no_taskmgr (false)
    , _canvas_type ("gdi")
#else
    , _canvas_type ("cairo")
#endif
{
    memset(_encrypt_channels, 0, sizeof(_encrypt_channels));

    parse_hotkeys("toggle-fullscreen=shift+f11,release-cursor=shift+f12"
#ifdef RED_DEBUG
                  ",connect=shift+f5,disconnect=shift+f6"
#endif
    );
}

void SpiceConfiguration::parse_hotkeys(const std::string& hotkeys)
{
    typedef boost::char_separator<char> separator;
    typedef boost::tokenizer<separator, std::string::const_iterator, std::string> tokenizer;

    LOG_DEBUG("hotkeys = " << hotkeys);

    separator sep(",");
    tokenizer criterias(hotkeys, sep);

    _hotkeys.clear();

    for (tokenizer::iterator iter = criterias.begin(); iter != criterias.end(); ++iter) {
        add_hotkey(*iter);
    }
}

void SpiceConfiguration::parse_keys(int action, std::string& hotkey)
{
    typedef boost::char_separator<char> separator;
    typedef boost::tokenizer<separator, std::string::const_iterator, std::string> tokenizer;

    separator sep("+");
    tokenizer criterias(hotkey, sep);

    std::vector<hotkey_t>& keys = _hotkeys[action];

    for (tokenizer::iterator iter = criterias.begin(); iter != criterias.end(); ++iter) {
        add_hotkey_redkey(keys, iter->c_str());
    }
}

int SpiceConfiguration::action_name(const std::string& action)
{
    if (strcasecmp(action.c_str(), "toggle-fullscreen") == 0) {
        return ActionToggleFullScreen;
    } else if (strcasecmp(action.c_str(), "release-cursor") == 0) {
        return ActionReleaseCursor;
    } 
#ifdef RED_DEBUG
    else if (strcasecmp(action.c_str(), "connect") == 0) {
        return ActionConnect;
    } else if (strcasecmp(action.c_str(), "disconnect") == 0) {
        return ActionDisconnect;
    } 
#endif // RED_DEBUG
    LOG_ERROR("Unknown action name " << action);
    return ActionUnknown;
}

void SpiceConfiguration::add_hotkey(const std::string& hotkey)
{
    typedef vector<string> split_vector_t;

    std::string action;
    std::string keys;

    vector<string> tokens;
    boost::split(tokens, hotkey, boost::is_any_of("="));

    if (tokens.size() == 2) {
        int action = action_name(tokens[0]);
        if (action != ActionUnknown) {
            LOG_DEBUG("action = " << tokens[0]);
            parse_keys(action, tokens[1]);
        }
    } else {
        throw Exception(fmt("%s: unable to parse hot keys") % __FUNCTION__);
    }
}

void SpiceConfiguration::add_hotkey_redkey(std::vector<hotkey_t>& keys, const char* key)
{
    ASSERT(key != NULL);

    static const struct {
        const char* name;
        RedKey main;
        RedKey alter;
    } keyboard[] = {
        { "alt", REDKEY_R_ALT, REDKEY_L_ALT },
        { "ralt", REDKEY_R_ALT, REDKEY_INVALID },
        { "rightalt", REDKEY_R_ALT, REDKEY_INVALID },
        { "right-alt", REDKEY_R_ALT, REDKEY_INVALID },
        { "lalt", REDKEY_L_ALT, REDKEY_INVALID },
        { "leftalt", REDKEY_L_ALT, REDKEY_INVALID },
        { "left-alt", REDKEY_L_ALT, REDKEY_INVALID },
        { "ctrl", REDKEY_R_CTRL, REDKEY_L_CTRL },
        { "rctrl", REDKEY_R_CTRL, REDKEY_INVALID },
        { "rightctrl", REDKEY_R_CTRL, REDKEY_INVALID },
        { "right-ctrl", REDKEY_R_CTRL, REDKEY_INVALID },
        { "lctrl", REDKEY_L_CTRL, REDKEY_INVALID },
        { "leftctrl", REDKEY_L_CTRL, REDKEY_INVALID },
        { "left-ctrl", REDKEY_L_CTRL, REDKEY_INVALID },
        { "shift", REDKEY_R_SHIFT, REDKEY_L_SHIFT },
        { "rshift", REDKEY_R_SHIFT, REDKEY_INVALID },
        { "rightshift", REDKEY_R_SHIFT, REDKEY_INVALID },
        { "right-shift", REDKEY_R_SHIFT, REDKEY_INVALID },
        { "lshift", REDKEY_L_SHIFT, REDKEY_INVALID },
        { "leftshift", REDKEY_L_SHIFT, REDKEY_INVALID },
        { "left-shift", REDKEY_L_SHIFT, REDKEY_INVALID },
        { "cmd", REDKEY_RIGHT_CMD, REDKEY_LEFT_CMD },
        { "rcmd", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "rightcmd", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "right-cmd", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "lcmd", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "leftcmd", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "left-cmd", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "win", REDKEY_RIGHT_CMD, REDKEY_LEFT_CMD },
        { "rwin", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "rightwin", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "right-win", REDKEY_RIGHT_CMD, REDKEY_INVALID },
        { "lwin", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "leftwin", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "left-win", REDKEY_LEFT_CMD, REDKEY_INVALID },
        { "esc", REDKEY_ESCAPE, REDKEY_INVALID },
        { "escape", REDKEY_ESCAPE, REDKEY_INVALID },
        { "ins", REDKEY_INSERT, REDKEY_INVALID },
        { "insert", REDKEY_INSERT, REDKEY_INVALID },
        { "del", REDKEY_DELETE, REDKEY_INVALID },
        { "delete", REDKEY_DELETE, REDKEY_INVALID },
        { "pgup", REDKEY_PAGEUP, REDKEY_INVALID },
        { "pageup", REDKEY_PAGEUP, REDKEY_INVALID },
        { "pgdn", REDKEY_PAGEDOWN, REDKEY_INVALID },
        { "pagedown", REDKEY_PAGEDOWN, REDKEY_INVALID },
        { "home", REDKEY_HOME, REDKEY_INVALID },
        { "end", REDKEY_END, REDKEY_INVALID },
        { "space", REDKEY_SPACE, REDKEY_INVALID },
        { "enter", REDKEY_ENTER, REDKEY_INVALID },
        { "tab", REDKEY_TAB, REDKEY_INVALID },
        { "f1", REDKEY_F1, REDKEY_INVALID },
        { "f2", REDKEY_F2, REDKEY_INVALID },
        { "f3", REDKEY_F3, REDKEY_INVALID },
        { "f4", REDKEY_F4, REDKEY_INVALID },
        { "f5", REDKEY_F5, REDKEY_INVALID },
        { "f6", REDKEY_F6, REDKEY_INVALID },
        { "f7", REDKEY_F7, REDKEY_INVALID },
        { "f8", REDKEY_F8, REDKEY_INVALID },
        { "f9", REDKEY_F9, REDKEY_INVALID },
        { "f10", REDKEY_F10, REDKEY_INVALID },
        { "f11", REDKEY_F11, REDKEY_INVALID },
        { "f12", REDKEY_F12, REDKEY_INVALID }
    };

    int i;

    for (i = 0; i < (sizeof(keyboard)/sizeof(keyboard[0])); ++i) {
        if (strcasecmp(key, keyboard[i].name) == 0) {
            hotkey_t hotkey;
            hotkey.main = keyboard[i].main;
            hotkey.alter = keyboard[i].alter;
            LOG_DEBUG("keys = " << keyboard[i].name);
            keys.push_back(hotkey);
            break;
        }
    }

    if (i == (sizeof(keyboard)/sizeof(keyboard[0]))) {
        LOG_ERROR("unknown key name: " << key);
    }
}

void SpiceConfiguration::parse_host_cert_subject(const std::string& subject)
{
    std::string::const_iterator iter = subject.begin();
    std::string entry;

    _host_cert_subject.clear();

    while (1) {
        if ((iter == subject.end()) || (*iter == ',')) {
            CertFieldValuePair entry_pair;
            int value_pos = entry.find_first_of('=');
            if ((value_pos == std::string::npos) || (value_pos == (entry.length() - 1))) {
                throw (fmt("host_subject bad format: assignment for %s is missing") %
                       entry.c_str());
            }
            entry_pair.first = entry.substr(0,value_pos);
            entry_pair.second = entry.substr(value_pos + 1);
            _host_cert_subject.push_back(entry_pair);
            DBG(0, fmt("subject entry: %s=%s") % entry_pair.first % entry_pair.second);
            if (iter == subject.end()) {
                break;
            }
            entry.clear();
        } else if (*iter == '\\') {
            iter++;
            if (iter == subject.end()) {
                LOG_WARN("single \\ in host subject");
                entry.append(1, '\\');
                continue;
            } else if ((*iter == '\\') || (*iter == ',')) {
                entry.append(1, *iter);
            } else {
                LOG_WARN("single \\ in host subject");
                entry.append(1, '\\');
                continue;
            }
        } else {
            entry.append(1, *iter);
        }
        iter++;
    }
}
