#include <sys/types.h>
#include <sys/socket.h>
#include <sys/epoll.h>
#include <sys/un.h>
#include "named_pipe.h"
#include "utils.h"
#include "debug.h"

int Session::accept(int fd_client, int epoll)
{
    struct epoll_event event;
    int flags;

    _fd_client = fd_client;
    _epoll = epoll;
    if ((flags = fcntl(fd_client, F_GETFL)) == -1) {
        LOG_WARN("accept failed, errno = " << errno << ", " << strerror(errno));
        return -1;
    }
    if (fcntl(fd_client, F_SETFL, flags | O_NONBLOCK) == -1) {
        LOG_WARN("accept failed, errno = " << errno << ", " << strerror(errno));
        return -1;
    }
    event.events = EPOLLIN | EPOLLOUT | EPOLLET;
    event.data.ptr = this;
    if (epoll_ctl(_epoll, EPOLL_CTL_ADD, _fd_client, &event) == -1) {
        LOG_WARN("epoll ctl failed, errno = " << errno << ", " << strerror(errno));
        return -1;
    }
    return 1; 
}

void Session::action(uint32_t event)
{
    _conn_interface->on_data();
}


void Session::bind(NamedPipe::ConnectionInterface* conn_interface)
{
    _conn_interface = conn_interface;
}

Session::~Session()
{
    struct epoll_event event; // BUG  in kernel versions before 2.6.9, can't send null

    if (epoll_ctl(_epoll, EPOLL_CTL_DEL, _fd_client, &event) == -1) {
        LOG_WARN("epoll ctl failed, errno = " << errno << ", " << strerror(errno));
    }
    close(_fd_client);
}

int32_t Session::write(const uint8_t *buf, int32_t size)
{
    const uint8_t *pos = buf;

    while (size) {
        int now;
        if ((now = send(_fd_client, (char *)pos, size, 0)) == -1) {
            if (errno == EAGAIN) {
                break;
            }
            if (errno == EINTR) {
                continue;
            }
            LOG_DEBUG("send error errno=" << errno);
            return -1;
        }
        size -= now;
        pos += now;
    }
    return (pos - buf);
}

int32_t Session::read(uint8_t *buf, int32_t size)
{
    uint8_t *pos = buf;
    while (size) {
        int now;
        if ((now = recv(_fd_client, (char *)pos, size, 0)) <= 0) {
            if (now == 0) {
                LOG_DEBUG("read error, connetion shutdown");
                return -1;
            }
            if (errno == EAGAIN) {
                break;
            }
            if (errno == EINTR) {
                continue;
            }
            LOG_DEBUG("read error errno=" << errno);
            return -1;
        }
        size -= now;
        pos += now;
    }
    return (pos - buf);
}

int LinuxListener::create_socket(const char *socket_name)
{
    int flags; 
    int listen_socket;
    struct sockaddr_un local;

    if ((listen_socket = socket(AF_UNIX, SOCK_STREAM, 0)) == -1) {
        LOG_DEBUG("create socket error, " << errno);
        return -1;
    }
    if ((flags = fcntl(listen_socket, F_GETFL, 0)) < 0) 
    { 
        LOG_DEBUG("get socket flags error, " << errno);
        return -1;
    } 
    if (fcntl(listen_socket, F_SETFL, flags | O_NONBLOCK) < 0) 
    { 
        LOG_DEBUG("set socket to non-blocking mode error, " << errno);
        return -1;
    } 

    _name = socket_name;

    local.sun_family = AF_UNIX;
    strcpy(local.sun_path, socket_name);
    unlink(local.sun_path);
    if (bind(listen_socket, (struct sockaddr *)&local,
             strlen(local.sun_path) + sizeof(local.sun_family)) == -1) {
        LOG_DEBUG("bind error, " << errno);
        return -1;
    }
    if (listen(listen_socket, 10) == -1) {
        LOG_DEBUG("listen error, " << errno);
        return -1;
    }
    return listen_socket;
}

LinuxListener::LinuxListener(const char *name, NamedPipe::ListenerInterface &listener_interface, int epoll) :
    _listener_interface(listener_interface), _epoll(epoll)
{
    struct epoll_event event;

    _listen_socket = create_socket(name);
    if (_listen_socket <= 0) {
        throw Exception(fmt("%s: Listener creation failed %n") % __FUNCTION__ % _listen_socket);
    }
    event.events = EPOLLIN;
    event.data.ptr = this;
    if (epoll_ctl(_epoll, EPOLL_CTL_ADD, _listen_socket, &event) == -1) {
        throw Exception(fmt("%s: epoll add fd failed %s") % __FUNCTION__ % strerror(errno));
    }
    LOG_DEBUG(fmt("listening socket - %s, added to epoll") % name);
}



LinuxListener::~LinuxListener()
{
   struct epoll_event event;
   if (epoll_ctl(_epoll, EPOLL_CTL_DEL, _listen_socket, &event) == -1) {
       LOG_WARN("epoll ctl failed, errno = " << errno << ", " << strerror(errno));
   }
   close(_listen_socket);
   unlink(_name.c_str());
}

void LinuxListener::action(uint32_t events)
{
    int fd_client;
    Session *conn;
    struct sockaddr_un remote;
    socklen_t len;

    len = sizeof(remote);
    if ((fd_client = accept(_listen_socket, (struct sockaddr *)&remote, &len)) == -1) {
        LOG_WARN("accept failed, errno = " << errno << ", " << strerror(errno));
        return;
    }
    conn = new Session();
    if (conn->accept(fd_client, _epoll) == -1) {
        delete conn;
        return;
    }
    LOG_DEBUG("New connection created, fd: " << fd_client);
    NamedPipe::ConnectionInterface &conn_interface = _listener_interface.create();
    conn->bind(&conn_interface);
    conn_interface.bind((NamedPipe::ConnectionRef)conn);
}
